/**********************-         bc.c       -*******************************
					   programable calculator
					   for the  command  line
					   by  John Venable  1991
		  Copyright 1991 John Venable - all rights reserved.

		  This program may be given away freely, but if you
		  start selling it you will have to pay me big bucks,
		  and I am talking serious coin here. ( If you wish
		  to contribute to the Venable Fund for underpaid
		  programmers you may send checks payable to me
		  at the address following this comment)

		This is a programable calculator for the command line.
		It can add, subtract, multiply, divide and raise to
		a power. It also can perform many interesting functions, 
		and run simple programs to boot. The result of its last
		operation is stored in a file (~~~~bc~~.~~~) in your
		temp directory ( or whatever is pointed to by the
		TEMP or TMP system variables) if it exists or in
		the current directory. When bc is executed again,
		the old result is loaded and can be recalled by the
		mem function.

		One day I was doing a few calculations at the computer
		and found myself having to pull out my pocket calculator
		to do them ( *yes*, I am a nerd who carries a calculator
		in my shirt pocket). Thinking to myself, why, with all
		this 386 iron sitting here, am I forced to rely on
		a puny calculator to do this? ( an original thought...)
		Why there should be a program to do this for me! Now, I
		know there are MANY pop-up calculators and such that I
		could have got for a song but why do things the easy way?
		I didn't buy this computer to make my life easy, I bought
		it to hack with! < maniacal laughter >! And, loving the UNIX
		approach of small tools combining to make really incomprehensible
		command lines, I decided that my command line calculator
		should also accept input from stdin. ( Shouldn't everyone? To
		heck with GUI's.) Then I started to get carried away. Why 
		should I hack up just a calculator when I could create a
		miniature programming language as well? ( You were wanting
		another programming language to play with, weren't you?)
		Well, this just barely qualifies as a programming language.
		The language supports these keywords and operators:

					   input  ( with optional prompt string)
					   print  ( string literals, variables, and expressions)
					   if
					   endif
					   goto
					   gosub
					   return
					   end
					   while
					   endwhile
					   shell   (to DOS)
					   dump    (variables)
					   load    (variables)
					   save    (variables)
					   mem     (result of last calculation)


					   = ( for assignment and comparison )
					   != 
					   <
					   >
					   >=
					   <=
					   +
					   - ( subtraction and negation )
					   *
					   /

		The language supports two data types - floating point doubles and
		arrays of floating point doubles.
		Goto's goto a label indicated by a trailing colon as in C.
		There are no string operations other than printing string literals.
		Variables can be up to 80 chars long - case sensitive.
		The built in functions can be 30 chars long and are case insensitive.

		My next project is writing a UNIX clone in my new language. ;>

		There is method in my madness, however. A calculator language
		with only six keywords is going to be easy to use. If I want
		complicated, I'll fire up C.
		

		One useful ( at least for me ) property of the program
		is its ability to add a list of numbers from stdin. If 
		you use the -a command line switch, the result of each
		line parsed is totaled. When this switch is in effect,
		the result of each expression is not displayed. Also, if
		two numbers are separated by only a space the program assumes
		that you wish to add them, not multiply them as you might expect.
		This is for me a feature (really!) not a bug. But again, feel
		free to change it. When in add mode you enter a equals
		sign '=' the program will respond with the subtotal of the 
		numbers you have entered. If you enter '==' (two equal signs)
		then the program will give you the grand total and clear
		the number stack to zero allowing you to start fresh.

		To add a new function all that is required is to #define the
		function's name with a unique value, add a if statement testing
		for the function in is in the what_function function, add a case
		statement in clear_functions, and of course add the function
		definition.

		The program works like this:
		The command line is checked to see if we are reading
		a program file or taking input from the command line or
		stdin. If it is a program file the lines to a maximum of
		MAX_PROGRAM are malloc'd into the statement array. The
		input lines are then  parsed first checking
		for exponents. They are cleared by adding parenthesis around
		them to make sure they are evaluated in the proper order. Then
		the parentheses are cleared. Then the functions are parsed and
		cleared. If it is not a function name it must be a variable so
		the variables are either assigned a value or a value is retrieved
		from them. A program function might not produce output, as
		a variable assignement will not.
		Then all multiplication and division is done with all
		results pushed onto the stack. The final step is to add all
		numbers on the stack and then push the result back onto it. This
		then is the final result. If input is from stdin or a program file,
		we loop until we meet up with eof or an end statement.

		Some changes you might want to do if so inclined...
			1. The program now just exits when an error is
			   detected. An error handling routine might be nice.
			2. The number of decimal places is controled by the
			   sprintf function. Changing the format from
			   "%lf" to "%0.9" will increase the number of
			   decimals. The accuracy of the right most digits
			   depends on the compiler's library. QuickC and
			   TurboC produced different results when I did that.

		I have compiled this program under TurboC++ and QuickC with
		no problems. I have also compiled it under Coherent with a
		few minor modifications. ( The changes for compiling under
		Coherent are:
			1... rename it calc.c ( or whatever) to
			     avoid conflicts with the venerable
				 bc.)
			2... use fixstack +4000 on the executable
			     to increase the stack size.

		If you do something interesting modifying this program,
		or find any (yikes!) bugs, I would like to see it. I have
		tested this program with as many problems as I could think
		of ( But of course I don't <really> want to find any more
		bugs in my program so I am probably not doing the best job of it)
		So, if this program proves useful to you, feel free to send
		me several hundred dollars or whatever. If you don't, well just
		don't send me taunting letters telling me how useful the program
		is and how much you enjoy using it for free. ( Just kidding.. I
		don't mind praise at all) If you think the code is dreck, and
		only a brain dead ex-BASIC jockey could so mangle the C language
		just don't forget that I know where you live and if you ever
		want to see your dog alive again you will take back all those
		nasty things you just said.

		See you in the bit stream.

								John Venable
								975 Cosgrave
								Winnemucca, Nevada 89445
								on Compuserve 72717,3072


*/

/* define here what compiler you are using - under UNIX the COHERENT switch should work */

#define QC
/*#define TC*/
/* #define COHERENT */

#include <stdio.h>
#include <math.h>
#ifndef COHERENT
	#include <stdlib.h>
#endif
#ifdef COHERENT
	#include <sys/types.h>
#endif
#include <string.h>
#include <time.h>
#include <ctype.h>

#ifdef TC
	extern int _stklen = 16000; /* set the stack size in turbo c */
#endif

#define EXP_VAR              /* define this to cause variables to be treated */
                             /* the same as numbers in exponetiation         */
#define STACK_SIZE 100
#define MAX_FUNCTION_NAME 33 /* maximum length of a function name */
#define MAX_BUFF 256         /* maximum lenghth of input buffer */
#define MAX_VARS 300          /* maximum number of variables */
#define VAR_LEN 80           /* maximum lenght of variable name */
#define MAX_PROGRAM 300      /* maximum number of program lines */
#define MAX_LEVELS 20        /* maximum number of while levels */

/* if you want to add a new function you must first define it here */

#define MULTIPLY 1
#define DIVIDE 2
#define POWER 3

#define VARIABLE 0 /* not a function, but a variable */
#define SQR  1
#define _PI   2
#define SIN  3
#define COS  4
#define TAN  5
#define ABS  6
#define ACOS 7
#define ASIN 8
#define ATAN 9
#define CEIL 10
#define COSH 11
#define EXP  12
#define LN  13
#define LOG 14
#define RAND 15
#define SINH 16
#define TANH 17
#define INT  18
#define CEXP 19
#define LOAN 20
#define MOD  21
#define HYPOT 22
#define MEM  23
#define DEG  24
#define RAD  25
#define ROOT 26
#define FACT 27
#define HELP 28
#define FRACT 29
#define CSC  30
#define SEC  31
#define COT  32
#define RECIP 33
#define PURM 34
#define INPUT 35
#define PRINT 36
#define END 37
#define IF 38
#define ENDIF 39
#define GOTO 40
#define SAVE 41
#define LOAD 42
#define GOSUB 43
#define RETURN 44
#define WHILE 45
#define ENDWHILE 46
#define SHELL 47
#define DUMP 48
#define REMAIN 49

int p_stack[ STACK_SIZE + 1];       /* global program stack */
int p_sp = 1;                       /* global program stack pointer */
char *statement[ MAX_PROGRAM + 1];  /* global pointers to program lines */
int _pc;                            /* program counter */
double stack [STACK_SIZE + 1];		/* global stack */
int sp = 1;							/* global stack pointer */
double memory;						/* global memory to save last result */
int _no_result;                     /* global flag - print a result or not */
char temp_path[160];                /* path to the memory variable saved to disk */

struct                              /* variables struct */
{
 	char name[VAR_LEN + 1];
	double amount;
} _variable[ MAX_VARS];                   /* global variable array */




/* define the new functions returns here */

double get_result();
double pop();
double root();
double loan();
double integer();
double deg();
double rad();
double cexp();
double random_number();
double fact();
double fract();
double csc();
double sec();
double cot();
double recip();
double purm();
double retrieve();
double remain();
void dump();
void init_vars();
void error();
void add_stack();
void clear_exp();
double modulus();


/*************************************************************************/

void main( argc, argv)
int argc;
char *argv[];
{
	char buffer[MAX_BUFF + 2]; /* input buffer */
	int add_mode;     /* adding matching or calculator ? */
	int program_mode; /* is it a program file? */
	FILE *fp;
	FILE *memfp;
	double result;    /* result of the calculation */
#ifndef COHERENT
	char *temp_environ;
#endif

	program_mode = 0;

	fp = stdin;
	add_mode = 0;     /* initialize them to zero */
	memory = 0;

/********************* check command line switches ************************/
	if( argc > 1)
	{
  		if( (argv[1][0] == '-' || argv[1][0] == '/') && argv[1][1] == '?' )
  		{
  			/* display usage screen */
  			usage();
  			exit(1);
  		}

  		if( argv[1][0] == '-' && argv[1][1] == 'a' )
  		{
  			/* set adding matching mode */
  			add_mode = 1;
		}
		if( argv[1][0] == '-' && argv[1][1] == 'f' )
		{
			/* get program from file */
			program_mode = 1;
			if( ( fp = fopen( argv[2], "r")) == NULL)
				error("error in opening file ", argv[2]);
		}

	}

/********************** get temp directory to store results ****************/
#ifndef COHERENT
	/* now get the path for the mem file  */
	temp_environ = getenv( "TMP"); /* the path to the temp subdirectory */
	if( temp_environ == NULL)
	{
		temp_environ = getenv( "TEMP"); /* try again */
	}
	strcpy( temp_path, temp_environ); /* if there is no temp set this is then a NULL string */
	if( temp_path == NULL)
	{
	 	strcpy( temp_path,"~~~~bc~~.tmp"); /* use the current directory */
	}
	else
	{
	 	strcat( temp_path,"\\~~~~bc~~.tmp"); /* add it to the path */
	}
#endif
#ifdef COHERENT
	strcpy( temp_path, "CALCMEMORY.TMP");
#endif
	if( ( memfp= fopen( temp_path, "rb")) == NULL)
	{
	 	memory = 0.0;                              /* doesn't exist */
	}
	else
	{
	 	fread( &memory, sizeof( double), 1, memfp); /* exists so read it */
		fclose( memfp);
	}

/*********************** initialize some globals ****************************/
	buffer[0] = '\0';   /* not a global but lets just be safe */
	init_vars();		/* initialize variable table */
	init_stack();       /* zero out the stack */

/*****************************| get terms from command line |****************/
	if( argc > 1 && ! add_mode && ! program_mode) 	/* get terms from command line */
	{
		--argc;
		++argv;
		while( argc--)
		{
			/* concatenate the command line args to one string */
			strcat( buffer, *argv++);
			strcat( buffer, " ");
			if( strlen( buffer) > ( MAX_BUFF /2) )
				error( "input too long..." , buffer);
		}
		_pc =2; /* this is for error messages using the program statement buffer */
		statement[1] = buffer;

		clear_exp(buffer);                   /* parenthesize exponents */
		clear_paren(buffer);                 /* clear the parentheses */
		result = get_result(buffer);      /* get results */
		
        if( ( result < 1e10 && result > 1e-9) || result == 0.0)
			sprintf( buffer,"%.9lf", result);/* format the results */
		else
			sprintf( buffer,"%.9lg", result);/* format the results */

		remove_tail( buffer);             /* get rid of trailing zeros */
		printf("%s\n", buffer);           /* shove it out to stdout */
		memory = result;
		save_memory();                    /* save the result to disk */

	}
	else                                  /* get terms from stdin */
	{

/*********************** store program in program buffer *******************/

		_pc = 0; 	/* program counter */
		if( fp != stdin ) /* program from program file */
		{
			while( fgets( buffer, (MAX_BUFF / 2), fp) != NULL)
			{
				buffer[strlen(buffer) -1] ='\0';/* get rid of new line */
				kill_remarks( buffer); /* get rid of the remarks by setting a \' to a \0 */
				if(strlen( buffer) < 1)         /* just hit enter so ignore it */
					continue;
				statement[_pc] = (char *) malloc( strlen( buffer) + 1);
				if( statement[_pc] == NULL)
					error(" out of memory - last statement was", statement[_pc -1]);
				kill_leading_space( buffer);
				strcpy( statement[_pc], buffer);
				++_pc;
			}
			statement[_pc] = NULL;   /* terminate the end of the program statements */
			fp = stdin;
			_pc = 0;
		}

/************* loop while input from stdin or program buffer ****************/
		for( ; ; )
		{
			/* loop while there is input */
			if( statement[_pc] == NULL && !program_mode)
			{

			/******** input from stdin, not from program buffer **********/

			 	if( fgets( buffer, (MAX_BUFF / 2), fp) == NULL)
					break;


				_pc =2; /* this is for error messages using the program statement buffer */

				statement[1] = buffer;

				_no_result = 0;                 /* set global result flag */
				buffer[strlen(buffer) -1] ='\0';/* get rid of new line */
				if(strlen( buffer) < 1)         /* just hit enter so ignore it */
					continue;
				if( buffer[0] == '?')             /* print help screen */
				{
					usage();
					continue;
				}

		/******************* check for add mode = functions *****************/

				if( add_mode && buffer[0] == '=' && buffer[1] == '=')
				{
					/* in add mode hitting '==' prints out total and clears stack */
					result = pop();
					push(0.0);                     /* zero out the stack */
					printf("_____________________________\n");

                    if( ( result < 1e10 && result > 1e-9) || result == 0.0)
						sprintf( buffer,"%.9lf", result);/* format the results */
					else
						sprintf( buffer,"%.9lg", result);/* format the results */

					remove_tail( buffer);
					strcat( buffer, "  Total");
					printf("%s\n", buffer);
					continue;
					}
	
					if( add_mode && buffer[0] == '=')
					{
						/* in add mode '=' prints out subtotal but does nothing to the stack */
						result = pop();
						push( result);
						printf("_____________________________\n");
						remove_tail( buffer);

                        if( ( result < 1e10 && result > 1e-9) || result == 0.0)
							sprintf( buffer,"%.9lf", result);/* format the results */
						else
							sprintf( buffer,"%.9lg", result);/* format the results */

						remove_tail( buffer);
						strcat( buffer, "  Subtotal");
						printf("%s\n", buffer);
						continue;
					}
			}/* end if */
			else
			{

			/**************** end of program so exit to DOS *****************/
				if ( statement[_pc] == NULL && program_mode)
				{
			 	save_memory();
				exit(0);
				}
			}

			/**************** loop while program buffer has statements *******/
			while( statement[_pc] != NULL)
			{
				strcpy( buffer, statement[_pc] );
				++_pc;
				_no_result = 0;                 /* set global result flag */

				clear_exp(buffer);                   /* parenthesize exponents */
				clear_paren(buffer);                 /* clear the parentheses */
				result = get_result(buffer);         /* get the results */
				if( _no_result)
					continue;
				memory = result;                     /* put the result in the memory */

				/* print out the results */

                if( ( result < 1e10 && result > 1e-9) || result == 0.0)
					sprintf( buffer,"%.9lf", result);/* format the results */
				else
					sprintf( buffer,"%.9lg", result);/* format the results */

				remove_tail( buffer);
				printf("%s\n", buffer);

			}/* end of while in program mode */


			/********************* get results from stdin input **************/
			clear_exp(buffer);                   /* parenthesize exponents */
			clear_paren(buffer);                 /* clear the parentheses */
			result = get_result(buffer);         /* get the results */

			if( _no_result)
				continue;
			memory = result;                     /* put the result in the memory */
			if( ! add_mode)
			{

				/* print out the results */

                if( ( result < 1e10 && result > 1e-9) || result == 0.0)
					sprintf( buffer,"%.9lf", result);/* format the results */
				else
					sprintf( buffer,"%.9lg", result);/* format the results */

				remove_tail( buffer);
				printf("%s\n", buffer);
			}
			/************** add mode - just push the results *****************/
			if( add_mode)
				push( result);


		}/* end of while loop */
		/************************* end of the input loop *********************/

		/********************* print out final tally if in add mode ***********/
		if( add_mode)
		{
		/* if in add mode, don't print it out until all input is finished */
			result = pop();


            if( ( result < 1e10 && result > 1e-9) || result == 0.0)
				sprintf( buffer,"%.9lf", result);/* format the results */
			else
				sprintf( buffer,"%.9lg", result);/* format the results */


            remove_tail( buffer);
			printf("%s\n", buffer);
		}
		save_memory(); /* save the memory variable */
		exit(0);
	}

}
/***************************************************************************/
init_stack()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: stack                                                   *
	 * purpose: be sure the stack has all zeros in it                   *
	 ********************************************************************/

 	int n;

	/* make it all zeros */
	for( n = 0; n < STACK_SIZE; n++)
		stack[ n ] = 0.0;
	return(0);
}
/***************************************************************************/
get_array( buff)
char *buff;
{
	/********************************************************************
	 * arguments: pointer to a string                                   *
	 * returns: nothing meaningful.                                     *
	 * globals: none.                                                   *
	 * purpose: to translate an array subscript to a number and copy    *
	 * it back into the passed string array                             *
	 ********************************************************************/

	int n,f;
	char var[ MAX_BUFF + 1];
	char subscript[80];
	double result;
	char *p;

	p = buff;
	n = 0;
	f = 0;
	strcpy( var, buff);

	/* gets array name and makes subscript a number, if not already */
	while( *buff != '[' && *buff != '\0')
	{
		++n;
	 	++buff;
	}
	if( *buff == '\0')
		error( "error in array ", statement[_pc-1]);

	/* n now points to the '[' char in var */
	var[ n + 1] = '\0'; /* terminate var after the '[' */
	++buff; /* skip the bracket */
	while( *buff != '\0' && *buff != ']')
	{
	 	subscript[f] = *buff;
		++f;
		++buff;
	}
	if( *buff == '\0')
		error( "error in array ", statement[_pc-1]);
	++buff;
	subscript[f] = '\0';

	result = get_result( subscript); /* convert any variables or functions to a number */
	sprintf(subscript, "%d", (int) result); /* convert that to an integer */
	strcat( var, subscript); /* place it at the end of the string */
	strcat( var, "]");       /* put the closing bracket on  - this prevents */
							 /* having more than a one dimensional array    */
	strcat( var, buff); /*  tack on whatever followed the array */
	buff = p; /* reset buff to where it was when we got it here */
	strcpy( buff, var);

	return(0);
}
/***************************************************************************/
kill_remarks( buff)
char *buff;
{
	/********************************************************************
	 * arguments: string pointer                                        *
	 * returns: nothing meaningful.                                     *
	 * globals: none.                                                   *
	 * purpose: kill a line following an apostrophy by making ' = \0    *
	 ********************************************************************/

	char *buffer;
	buffer = buff;

	/* get rid of comments and leading blanks in buff */
	while( isspace( *buffer) )
		++buffer;
	strcpy( buff, buffer);

	while( *buff != '\0')
	{
	 	if( *buff == '\"')
		{
		 	++buff;
			while( *buff != '\"' && *buff != '\0') /* skip quotes */
				++buff;
		}
		if( *buff == '\'') /* is it the apostrophy? */
		{
		 	*buff = '\0';  /* just kill it and what follows it */
			return(0);
		}
		++buff;
	}

	return(0);
}
/***************************************************************************/

save_memory()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: memory                                                  *
	 * purpose: save the global memory to disk upon exit from program   *
	 ********************************************************************/

FILE *fp;
	if( ( fp = fopen( temp_path, "wb")) == NULL)
		return(0); /* no panic, just won't be able to save the result memory to disk */
	else
		fwrite( &memory, sizeof( double), 1, fp);

	fclose(fp);
	return(0);
}

/***************************************************************************/
usage()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: print help screen.                                      *
	 ********************************************************************/

	printf(" %s\n %s\n %s\n %s\n %s\n %s\n ",
	"A Command Line Scientific Calculator by John Venable 1991", 
	"Usage = bc [-?af <file.nam>]  [expression]....",
	"If no command line is given the program defaults to the standard input.",
	"-? - prints this message.",
	"-a - set add mode, for adding a list of numbers from stdin.",
	"-f <file.nam> runs the program contained in <file.nam>"
	);
	printf( " %s\n %s\n %s\n %s\n %s\n %s\n",
	"When in add mode \"=\" gives a subtotal and \"==\" gives a total and clear.",
	"save file.nam - save variables    | end - exits program",
	"load file.nam - load variables    | pi - return 3.1415...",
	"sin, cos, tan, csc, sec, cot      | rand - random number",
	"asin, acos, atan                  | ceil - next higher integer number",
	"sinh, cosh, tanh                  | sqr - square root"
	);
	printf(" %s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n",
	"exp - e to the xth power          | ln - natural logarithm",
	"cexp - 10 to the xth power        | log - common base 10 logarithm",
	"abs - absolute value              | loan - amount, interest, months = payment",
	"mod - x modulus y ( mod x y)      | hypot - a,b = hypotenuse of right triangle",
	"mem - result of last calculation  | deg - convert radians to degrees",
	"rad - convert degrees to radians  | root - xth root of y",
	"fact - factorial                  | help - prints this screen",
	"fract - fractional amount         | int - truncate decimal fraction",
	"recip - reciprocal                | purm - purmutation x, y       ",
	"shell - execute DOS command       | dump - dump variables         ",
	"remain - remainder x/y            |"
	);


return(0);
}

/***************************************************************************/
p_push( val)
int val;
{
	/********************************************************************
	 * arguments: integer val - normaly the program counter _pc         *
	 * returns: nothing meaningful.                                     *
	 * globals: p_stack, p_sp                                           *
	 * purpose: push the program counter for while loops                *
	 ********************************************************************/

	/* push numbers on the global program stack */
	p_stack[ p_sp] = val;
	++p_sp;
	if( p_sp > STACK_SIZE)
	{
		error(" program stack overflow - too many nested levels of gosubs or whiles","");
	}
	return(0);
}
/***************************************************************************/
p_pop()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: contents of the p_stack (program stack)                 *
	 * globals: p_stack, p_sp                                           *
	 * purpose: returns the contents of stack the decrements the sp.    *
	 ********************************************************************/

	/* pop program counter off the stack */
	if( p_sp < 1 )
		error (" program stack underflow", "");

	--p_sp;

	return( p_stack[p_sp] );
}
/***************************************************************************/
push( val )
double val;
{
	/********************************************************************
	 * arguments: double val                                            *
	 * returns: nothing meaningful.                                     *
	 * globals: stack, sp                                               *
	 * purpose: push double on the stack and inc the sp                 *
	 ********************************************************************/

	/* push numbers on the global stack */
	stack[sp] = val;
	++sp;
	if( sp > STACK_SIZE)
	{
		error(" stack overflow - term too complex", "");
	}
	return(0);
}
/***************************************************************************/
double pop()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: contents of the stack                                   *
	 * globals: stack, sp                                               *
	 * purpose: pop off the stack and dec the sp                        *
	 ********************************************************************/

	/* pop numbers off the stack */
 	if( sp < 1 )
		error (" stack underflow", "");

	--sp;

	return( stack[sp] );
}
/***************************************************************************/
swap()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: stack                                                   *
	 * purpose: exchange the top two members of the stack               *
	 ********************************************************************/

	/* swap stack members */
	double a, b;

	a = pop();
	b = pop();
	/* this is so screwball - why this is required I have no idea */
	#ifdef QC
		push(b);
		push(a);
	#endif
	#ifdef TC
		push(a);
		push(b);
	#endif
	#ifdef COHERENT
		push(b);
		push(a);
	#endif
	return(0);

}

/***************************************************************************/


void error( string, errorstring)
char *string;
char *errorstring;
{
	/********************************************************************
	 * arguments: string pointers string, errorstring                   *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: prints error message then exits to DOS                  *
	 ********************************************************************/

	/* print error then exit */
	fprintf(stderr, "%s  : %s\n", string, errorstring);
	exit(1);
}

/***************************************************************************/
isoperator( c)
int c;
{
	/********************************************************************
	 * arguments: char c                                                *
	 * returns: true or false                                           *
	 * globals: none                                                    *
	 * purpose: true if * or / else false                               *
	 ********************************************************************/

	/* is it multiply, divide or power? */
	if( c == '*' || c == '/' || c == '^')
		return(1);
	else
		return(0);
}
/***************************************************************************/

double get_result(buffer)
char *buffer;
{
	/********************************************************************
	 * arguments: string pointer buffer                                 *
	 * returns: double result of expression                             *
	 * globals: _no_result                                             *
	 * purpose: parse and solve an expression                           *
	 ********************************************************************/

	char *buff;                /* points to buffer */
	int mode;                  /* mult, div, or exp */
	int n;
	double result;
	double number;
	char var[VAR_LEN +2];      /* holds variable to assign results to, if any */
	char temp[MAX_BUFF + 1];


	var[0] = '\0';
	/* check if this is a linelabel */
	if( strchr( buffer, ':') != NULL)
	{
	 	if( ! strstr( buffer, "print") && ! strstr( buffer, "input") )
		{
			/* this is then a line label - do nothing */
		 	_no_result = 1;
			return(0);
		}
	}
	/* check to see if this is a variable assignment - but not if it is a print or a if statement */
	if( strchr( buffer, '=' ) && ! strstr( buffer, "print ") && ! strstr( buffer, "if ") && ! strstr( buffer, "input") )/* copy the assignment statement into var */
	{
		strcpy( temp, buffer);
		buff = temp;
		n = 0;

		while( isspace( *buff) )
			++buff;

		while(
				! isspace( *buff) &&
		        *buff != '='      &&
				*buff != '>'      &&
				*buff != '<'      &&
				*buff != '!'      &&
				*buff != '\0'
			 )
		{
		 	var[n] = *buff;

		 	if( isalpha( *buff) || *buff == '[' || *buff == '-')
			{
				if( *buff == '[')
				{
					get_array(buff);
					while( *buff != ']' && *buff != '\0')
					{
						var[n] = *buff;
						++buff;
						++n;
						if( n > VAR_LEN)
							error( "error in subscript", statement[_pc-1]);
					}
					if( *buff == '\0')
						error( "error in statement ", statement[_pc-1]);
					var[n] = *buff; /* this should tack on the bracket */
				}

				++n;
				if( n > VAR_LEN)
				{
					var[n] = '\0';
					error( "variable is too long", var);
				}
			}
 			++buff;
		}
		var[n] = '\0';
		if ( what_function( var) == IF || what_function(var) == WHILE)
		{
			buff = temp;
			var[0] = '\0';
		}
		strcpy( buffer, buff); /* debug - was ++buff */

	}
	buff = buffer;             /* initialize */
	
	clear_functions(buffer);   /* find and clear any functions */

	while( *buff != '\0')
	{
		while( ! isoperator( *buff ) && *buff != '\0') /* go until /,* or ^ or end */
		{
			if( isspace( *buff) )
			{
				/* skip any spaces */
				++buff;
				continue;
			}
			result = get_number( buff ); /* get a number from the buff and pop it on the stack */
		}

		if( strlen( buff) < 1)          /* if so, all done */
			break;

		if( *buff == '*')
		{
			mode = MULTIPLY; /* multiply */
			++buff;
		}

		if( *buff == '/')
		{
			mode = DIVIDE; /* divide */
			++buff;
		}

		if( *buff == '^')
		{
			mode = POWER; /* raise to a power */
			++buff;
		}

		get_number( buff ); /* get the other number and pop it on the stack */

		number = pop(); /* pop them off */
		result = pop();

		switch( mode )
		{

			case MULTIPLY:
				push( number * result );
			break;

			case DIVIDE:
				if( number == 0)
					error( " divide by zero ","");
				push( result / number );
			break;

			case POWER:
				push( pow(result, number));
			break;

			default:
				error(" Invalid operator ","");
			break;
		}/* end switch */

	}

	add_stack();   /* all multiplication, division, powers and functions have */
	               /* been cleared - now only addition remains of the numbers */
				   /* on the stack. Add them then push result.               */
	result = pop();/* pop it back off */
	buff = buffer;         /* reset buff */
	strcpy( buffer, buff); /* copy it back into the buffer */
	if( var[0] != '\0')
	{
		_no_result = 1;
		assign( var, result);
		return(0);
	}
	else
	{
		return( result);       /* that's it */
	}
}

/***************************************************************************/
assign( var, amount)
char *var;
double amount;
{
	/********************************************************************
	 * arguments: string pointer var, double amount                     *
	 * returns: nothing meaningful.                                     *
	 * globals: _no_result, _variable                                   *
	 * purpose: place a variable name and value in the variable table   *
	 ********************************************************************/

	int n;

	if( what_function( var) )
	{
		printf("Used reserved name for variable : %s\n", var); /* can't use a function name for a variable */
		return(0);
	}

	for( n = 0; n < MAX_VARS; n++)
	{
	 	if( strcmp( _variable[n].name, var) == 0)
		{
		 	_variable[n].amount = amount;
			break;
		}
		if( _variable[n].name[0] == '\0')
		{
		 	strcpy( _variable[n].name, var);
			_variable[n].amount = amount;
			break;
		}

	}
	if( n >= MAX_VARS)
	{
		printf("Too many variables  out of space : %s\n", var);
	 	_no_result = 1;
	}


	return(0);
}

/***************************************************************************/
double retrieve( var)
char *var;
{
	/********************************************************************
	 * arguments: string pointer var                                    *
	 * returns: value held in the variable table for var                *
	 * globals: _no_result, _variable                                   *
	 * purpose: search for the variable name and if found return value  *
	 *          if it is not found, add it and issue error message      *
	 ********************************************************************/

 	int(n);

	for( n = 0; n < MAX_VARS; n++)
	{
	 	if( strcmp( _variable[n].name, var) == 0)
		{
		 	return( _variable[n].amount);
		}
		if( _variable[n].name[0] == '\0' && n < MAX_VARS)
		{
			printf(" You are referencing a variable that has been assigned no value : %s\n", var);
			assign( var, 0.0);
			_no_result = 1;
			return(0.0);
		}
        if( _variable[n].name[0] == '\0' && n >= MAX_VARS)
		{
		 	printf( " No more room for another variable : %s\n", var);
			_no_result = 1;
			return(0.0);
		}

	}
	printf( " No more room for another variable : %s\n", var);
	_no_result = 1;
	return( 0.0);
}
/***************************************************************************/
void init_vars()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: _variable                                               *
	 * purpose: set all variable names to null and values to zero       *
	 ********************************************************************/

	int n;
	for( n = 0; n < MAX_VARS; n++)
	{
	 	_variable[n].name[0] = '\0';
		_variable[n].amount = 0.0;
	}
}
/***************************************************************************/

void add_stack()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: stack, sp                                               *
	 * purpose: add everything on the stack and then push the result    *
	 ********************************************************************/

	double result;

	/* add all of the numbers on the stack and push the result back on */

	result =0;

	if( sp < 1)
		error(" no result - stack empty","");
	while( sp > 1)
	{
	 	result += pop();
	}
	push( result );

}

/***************************************************************************/

get_number(buffer)
char *buffer;
{
	/********************************************************************
	 * arguments: string pointer buffer                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: parse a number from the buffer, convert it and push it  *
	 *          onto the stack.
	 ********************************************************************/

 	double result;   /* final result */
	int f;           /* loop index */
	char temp[128];  /* hold temp string */
	char *buff;      /* points to buffer */
	int sign;        /* positive or negative? */

	sign = 1;        /* start out positive */

	buff = buffer;   /* initialize buff and f */
	f = 0;

	while( isspace(*buff) || *buff == '+')    /* skip spaces */
		++buff;

	if( *buff == '-')
	{
		sign = (- sign);       /* minus so negate sign */
		++buff;
	}

	while( ! isdigit( *buff ) && *buff != '.' && *buff != '\0')
	{
		if( isalpha( *buff) )       /* alpha means a function */
		{
			clear_functions( buff); /* convert it to a number */
			break;
		}

		if( *buff == '-')
		{
			/* check for minus sign */
			sign = (- sign);
			++buff;
			continue;
		}
		if( *buff == '(')
		{
			clear_paren(buff);    /* get rid of the parentheses */
			break;
		}
		
		++buff;
	}
	if( *buff == '\0')
		error("error in expression",statement[_pc -1]);  /* oops... */

	while(
			isdigit( *buff ) ||
			*buff == '.'     ||
			*buff == 'e'     ||
			*buff == '-'     ||
			*buff == '+'
		 )
	{
		if( *buff == '-' || *buff == '+')
		{
			/* figure out negative or positive */
				if( *buff == '-' && f == 0)
				{
					sign = (- sign);
					++buff;
					continue;
				}
				else
					if( temp[f - 1] != 'e')
						break; /* not scientific notation */
		}
		temp[f] = *buff;
		++buff;
		++f;
		if( f > 45)
		{
			/* c math functions are only so accurate so no point in going overboard */
			temp[f] = '\0';
			error( " number too big  ", temp);
		}
	}
	temp[ f] = '\0';
	remove_tail( temp); /* get rid of trailing zeros */

	if( strlen( temp) > 45)
		error( "number too big ",temp);

	errno = 0;              /* global math error flag */
	result = atof( temp);   /* convert to double */
	result = result * sign; /* make it the right sign */

	if( errno == ERANGE)
		error("Number too large or too small","");

	push(result);           /* put it on the stack */
	strcpy( buffer, buff);  /* reset buffer with new value */
	return(0);

}
/**************************************************************************/

clear_paren(buff)
char *buff;
{
	/********************************************************************
	 * arguments: string pointer buff                                   *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: replaces the contents of parentheses with a number      *
	 *          and then replaces the string buff with the results      *
	 ********************************************************************/

	/* this function clears out the parentheses, starting with
	   the inner most ones.
	*/
	
	char temp[MAX_BUFF + 2];    /* buffers to hold intermediate results */
	char temp2[MAX_BUFF + 2];
	char buffer[MAX_BUFF + 2];
	int paren;         /* holds the level of parentheses */
	int n,f,i;         /* loop indices */
	int last_paren;    /* position of last parenthesis */
	double result;     /* result of the effort */


	paren = 0;         /* start with no parentheses */

	for(n = 0;; n++)   /* start looking for them */
	{
		if( buff[n] == '\"')
			return(0); /* this is a quoted string */

 		if( buff[n] == '\0')
		{
			if( paren)  /* whoops.. end of string and haven't cleared them all */
				error("error in parentheses....","");

			return(0);  /* finished */
		}
		
		if( buff[n] == '(')
		{
			++paren;          /* increment the level */
			last_paren = n;   /* this might be the last one, so mark it */
		}
		if( buff[n] == ')')
		{
			--paren;          /* bumped into the other end so decrement the level */

			for( f = last_paren + 1, i = 0; f < n; f++, i++)
			{
				buffer[i] = buff[f];   /* copy what's between the parens into the buffer */
			}

			strcpy( temp, buff);      /* store the original string */
			strcpy( temp2, buff);

			buffer[i] = '\0';          /* terminate the inside of the parens */
			strcpy(buff, buffer);      /* move the parens to buff */

			result = get_result(buff); /* get the result and put it into buff */

			/* sprintf the double result into
			   the position of the left paren */


            if( ( result < 1e10 && result > 1e-9) || result == 0.0)
				sprintf( &temp2[last_paren],"%.9lf", result);/* format the results */
			else
				sprintf( &temp2[last_paren],"%.9lg", result);/* format the results */


			strcat( temp2, &temp[n + 1]);   /* now, append the rest of the string that followed
			                                   the right paren onto the new string */
			strcpy( buff, temp2);           /* then copy the new string with the parenthetical
			                                   expression replaced by a single number back into
											   the original string */


			n = ( -1);                      /* then reset the loop index so we
			                                   will start the loop at the beginning
											   of the string to clear any remaining
											   parentheses. */
			paren = 0;                      /* reset the parentheses level to zero */

		}

	}



}
/**************************************************************************/
remove_tail( string )
char *string;
{
	/********************************************************************
	 * arguments: string pointer string                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: delete trailing zeros from string                       *
	 ********************************************************************/

	/* I remove the trailing zeros because they annoy me. */

	int n;
	n = strlen( string) - 1;

	if( strchr( string, 'e') || strchr( string, 'E')) /* if exponential format */
		return(0);                                    /* just do nothing       */

	if( ! strchr( string, '.') )                      /* if no decimal point   */
		return(0);                                    /* just do nothing       */


	while( string[n] == '0' || string[n] == '.' && n > 0)
	{
		/* get rid of trailing zeros and useless decimal points as well */

		if( string[n] == '.')
		{
			string[n] = '\0';
			return(0);
		}
		string[n] = '\0';

		--n;
	} 
	return(0);


}
/***************************************************************************/

void clear_exp(buffer)
char *buffer;
{
	/********************************************************************
	 * arguments: string pointer buffer                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: put parentheses around exponential expressions          *
	 ********************************************************************/

	/* for proper evaluation order exponentiation must be performed first.
	   I chose as the easiest way to do this was to just surround them
	   with parentheses and let clear_parens do the actual work. Here,
	   we just find the beginning and end and insert parens around them
	   and return the resulting string.
	*/

	int begin;         /* the start of the expression */
	int last;          /* the end of it */
	int n, f, i;          /* loop indices  */
	char temp[MAX_BUFF + 2];    /* temporary storage */
	char temp2[MAX_BUFF + 2];
	int paren;         /* parenthesis level */


	for( n = 0; n < strlen( buffer); n++)
	{
		/* start looking for the caret */

		if( buffer[n] == '^')
		{
 			if( n == 0)
				error("error in exponent", buffer);/* can't have it first without preceeding argument */

			for( f= n -1; f >= 0; f--)
			{
				if( ! isspace( buffer[f]))         /* discard spaces */
					break;
			}

			for( ; f >= 0; f--)     /* first find start of term to be raised */
			{                       /* set the level of parens to zero */
				paren = 0;
					
				if( buffer[f] == ')' )  /* if the number to be raised is a parenthetical expression */
				{
					++paren;            /* up the level of parentheses */
					--f;                /* go back one more char in the string */

					for( ;; --f)        /* now start searching for the start of the expression */
					{
						if( f < 0)      /* hit the start of the string but not the start of the expression */
							error("error in exponetial expression",buffer);

						if( buffer[f] == ')')
							++paren;    /* hit another level of parenthesis */

						if( buffer[f] == '(')
						{
							--paren;
							if( ! paren )
								break;  /* found the outer parenthesis so exit this loop */
						}
					}
					begin = f;          /* point begin to the start of the parenthisis */
					break;              /* exit the loop */
				}
				else                    /* no parentheses so just look for a number */
				{
					while( isspace( buffer[f] ))
					{
						--f;           /* skip the spaces */
						if( f < 0)     /* oops...         */
							error(" error in exponent","");
					}

					/* I used a infinite loop here because I wanted to use the
					   break statement to exit without executing any other nested
					   loops. This saved me a goto statement <whew>. This loop
					   is executed only once.
					*/
					for(;;)
					{
						if( isdigit( buffer[f]) || buffer[f] == '.') /* is it a digit or a decimal point?*/
						{
							for( ; isdigit( buffer[f]) || buffer[f] == '.' ; f--) /* if it is find the start of it */
							{
								if( f < 0)
									error("error in expression",buffer);
							}
					
							begin = f+1;     /* position of term to be raised to power */
							break;           /* this is why I used the for loop. a break here
												skips all of the following code in the loop */
						}

						if( isalpha( buffer[f]) ) /* not a number, so it must be a constant  or a variable*/
						{
							for( ; isalpha( buffer[f]); f--) /* find the start of it */
							{
								if( f < 0)
									error("error in expression",statement[_pc-1]);
							}
					
							begin = f+1;       /* position of term to be raised to power */
							for( i = begin; i < n; i++)
							{
								temp[ i - begin] = buffer[i];
							}
							temp[ i - begin] = '\0';
							clear_functions( temp);
							strcpy( temp2, buffer);
							strcpy( &temp2[ begin], temp);
							strcat( temp2, &buffer[n]);

							#ifndef EXP_VAR
							if( temp[0] == '-')   /* this causes variables to be treated the */
								begin = begin + 1;/* same as a number for exponentiation.    */
								                  /* if a = -10 then a^2 = -100 not 100      */
												  /* as would be correct in math.            */
							#endif

							strcpy( buffer, temp2);
							for( n = 0; buffer[n] != '^'; n++) ; /* find the ^ again */

							break;             /* skip the rest of the loop */
						}

						error( "error in exponent","");  /* else don't know what is here so exit with a error */
					}
					break; /* break out of the outer loop */
				}
			}/* end for */

			/* now find the exponent */
			for( f = n + 1; isspace(buffer[f]); f++) /* skip the spaces*/
			{
				if( f > strlen( buffer))
					error(" error - unexpected end of expression ","");								
			}

			last = 0;       /* the end of the exponent */
			paren = 0;      /* the levels of parentheses */

			while( buffer[f] == '-' || buffer[f] == '+')
				++f;        /* skip leading signs */
			
			for( ; f < strlen(buffer);)
			{

				if( buffer[f] == '(')              /* is the exponent in parentheses? */
				{
					++paren;                       /* increment the level of parentheses */
					++f;
					for( ;f < strlen( buffer);f++)
					{
						if( buffer[f] == '(')      /* inc the parentheses level */
							++paren;
						if( buffer[f] == ')' )     /* found a closing paren     */
						{
							--paren;               /* decrement the paren level */
							if( ! paren)           /* if paren == 0 we are done */
							{
								last = f;          /* last = the end of the parentheses */
								break;             /* exit from the loop */
							}
						}
			
					}
					if( f >= strlen( buffer))      /* never found the closing paren */
						error("error in parentheses","");
					else
						break;                     /* exit from the outer loop, we found the end */
					
				}

				/* no parentheses so it is either a number or a function */

				while( ! isdigit( buffer[f]) && ! isalpha( buffer[f]) && buffer[f] != '\0' )
					++f;	/* skip spaces and signs */

				if( isalpha( buffer[f]) )         /* if it is an alpha it must be a function or a constant */
				{
					clear_functions( &buffer[f]); /* so clear it */
				}
				++f;                              /* inc the loop index */

				/* if there was a function, it will have been converted
	               to a number and must now be parsed */

				for( ; isdigit( buffer[f]) || buffer[f] == '.'; f++)
				{
					if( f > strlen( buffer))
						error("Unexpected end of exponent","");
				}
				last = f;                         /* found the last of the exponent */
				break;                            /* exit the loop */
	
			}/* end of for */

			for( f = 0; f< begin; f++)
			{
				temp[f] = buffer[f];              /* copy the first part of the string */
			}

			temp[f] = '(';                        /* insert a left parenthesis */

			for( f = begin; f<= last; f++)
			{
				temp[f + 1] = buffer[f];          /* insert the exponential expression */
			}

			temp[f] = ')';                        /* append a right parenthesis */

			for( f = last; buffer[f] != '\0'; f++)
			{
				temp[f + 2] = buffer[f];          /* append the rest of the string */
			}
			temp[f + 2] = '\0';                   /* terminate it */

			strcpy( buffer, temp);                /* copy the temp back into the buffer */

			++n;                                  /* increment the index loop */


		}/* end if n = '^' */
	}



}

/*************************************************************************/

clear_functions(buffer)
char *buffer;
{
	/********************************************************************
	 * arguments: string pointer buffer                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: _no_result                                              *
	 * purpose: replace functions,their arguments, and variables with   *
	 *          their results.                                          *
	 ********************************************************************/

	/* find functions and convert them to numbers */

	char temp[MAX_BUFF + 2];     /* temporary string storage */
	char *buff;         /* points to buffer */
	int n, f;           /* loop indices */
	char function[MAX_FUNCTION_NAME + 2];  /* stores the function name */
	int funct;          /* holds the #define'd function number */
	int start;          /* the starting position of the function name */
	double arg;         /* the result of the function */


	buff = buffer;      /* initialize the pointer */

	strcpy(temp, buff); /* copy the buffer into temp  to save it */
	

	for( n = 0; *buff != '\0'; buff++, n++)
	{

		if( isalpha(*buff ))
		{
			f = 0;
			start = n;               /* beginning position of function name */
			function[f] = *buff; /* copy the function name into function */
			while( isalpha( *buff) || *buff == ':' || *buff == '[' )
			{
				if( *buff == '[')
					get_array( buff);

				function[f] = *buff; /* copy the function name into function */

				if( *buff == '[') /* this is an array so get the subscript*/
				{
					function[f] = *buff;
					++buff;
					++f;
					while( *buff != ']' && *buff != '\0')
					{
						function[f] = *buff;
					 	++buff;
						++f;
					}
					if( *buff == '\0')
						error( "error in array ", statement[_pc-1]);
					function[f] = *buff;
				}

				++n;
				++f;
				++buff;
				if( f > MAX_FUNCTION_NAME)
				{
					function[f] = '\0';
					error("error in function name ", function); /* function name too long */
				}
			}
			function[f] = '\0'; /* terminate the function name */
			if( strchr( function, ':') )
			{
				_no_result = 1;
				++buff;
				*buff = '\0';
				--buff;
				temp[0] = '\0';
				continue; /* this is a program label */
			}

			/* this code checks for scientific format */
			if( function[0] == 'e' && n > 1 )
			{
				--buff; /* check if preceeding the 'e' is a digit*/
				--buff;
				if( function[0] == 'e' && function[1] == '\0' && isdigit(*buff) )
				{
					++buff;
					++buff;
					continue;  /* this is scientific notation so skip it*/
				}
				++buff;
				++buff;
			}

			funct = what_function( function ); /* returns the #defined function number */

			errno = 0;                         /* set error global */

			switch( funct )
			{
				case VARIABLE:
					arg = retrieve( function); /* get the value held in the variable */
					insert( buff, temp, arg, start);
				break;

				case SQR:
					get_number(buff);          /* get_number pushes the number on the stack */
					arg = pop();
					if( arg < 0.0)
					{
					 	printf("domain error - you cannot take a square root of a negative number\n");
						_no_result = 1;
						break;
					}
					push( arg);
					arg = sqrt( pop());        /* then you pop it off for the function's argument */
					if( errno )                /* if errno is set some math error has occured */
						error( "error in ", function);
					insert( buff,  temp, arg, start);  /* insert puts the number into the string     */
				break;                                 /* in place of the function and it's argument */

				case _PI:
					/* PI is really a constant and should be handled seperately, but this is easier */
					arg = 3.141592654;
					insert( buff,  temp, arg, start);
				break;
				
				case SIN:
					get_number(buff);
					arg =  sin( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case COS:
					get_number(buff);
					arg = cos( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case CSC:
					get_number(buff);
					arg =  csc( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case SEC:
					get_number(buff);
					arg =  sec( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case COT:
					get_number(buff);
					arg =  cot( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case TAN:
					get_number(buff);
					arg = tan( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;
					
				case ABS:
					get_number(buff);
					arg = fabs( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case ACOS:
					get_number(buff);
					arg = acos( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case ASIN:
					get_number(buff);
					arg = asin( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case ATAN:
					get_number(buff);
					arg = atan( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case CEIL:
					get_number(buff);
					arg = ceil( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case COSH:
					get_number(buff);
					arg = cosh( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case EXP:
					get_number(buff);
					arg = exp( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case LN:
					get_number(buff);
					arg = log( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case LOG:
					/* this is common log base 10. I couldn't use log10() because it contained digits */
					get_number(buff);
					arg = log10( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case RAND:
					/* I used the random number routine from the book */
					/* The Most Popular Subroutines in BASIC by Ken Tracton */
					arg =  random_number();
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case SINH:
					get_number(buff);
					arg = sinh( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case TANH:
					get_number(buff);
					arg = tanh( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case INT:
					get_number(buff);
					arg = integer( pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case CEXP:
					get_number(buff);
					arg = cexp( pop() );
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case LOAN:
					/* this routine pops three numbers - so don't put parentheses around them */
					get_number(buff);
					get_number(buff);
					get_number(buff);
					arg = loan( pop(), pop(), pop() );
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case MOD:
					/* this routine pops two numbers - so don't put parentheses around them */
					get_number(buff);
					get_number(buff);
					arg = modulus( pop(), pop() );
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case REMAIN:
					/* this routine pops two numbers - so don't put parentheses around them */
					get_number(buff);
					get_number(buff);
					arg = remain( pop(), pop() );
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case HYPOT:
					/* this routine pops two numbers - so don't put parentheses around them */
					get_number(buff);
					get_number(buff);
					arg = hypot( pop(), pop());
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case MEM:
					/* this variable holds the result of the last calculation */
					arg = memory;
					insert( buff,  temp, arg, start);
				break;

				case DEG:
					/* radians to degrees */
					get_number(buff);
					arg = deg( pop() );
					insert( buff,  temp, arg, start);
				break;

				case RAD:
					/* degrees to radians */
					get_number(buff);
					arg = rad( pop() );
					insert( buff,  temp, arg, start);
				break;

				case ROOT:
					/* this routine pops two numbers - so don't put parentheses around them */
					get_number(buff);
					get_number(buff);
					arg = root( pop(), pop() );
					if( errno )
						error( "error in ", function);
					insert( buff,  temp, arg, start);
				break;

				case HELP:
					usage();
					arg = 0.0;
					insert( buff, temp, arg, start);
					_no_result = 1;
				break;
				
				case FACT:
					/* factorial */
					get_number(buff);
					arg = fact( pop() );
					insert( buff,  temp, arg, start);
				break;

				case FRACT:
					/* get fractional part of number */
					get_number(buff);
					arg = fract( pop() );
					insert( buff,  temp, arg, start);
				break;

				case RECIP:
					/* get reciprocal */
					get_number(buff);
					arg = recip( pop() );
					insert( buff,  temp, arg, start);
				break;

				case PURM:
					/* get purmutations */
					get_number( buff);
					get_number( buff);
					arg = purm( pop(), pop() );
					insert( buff, temp, arg, start);
				break;

				case INPUT:
					  get_input( buff);
					  _no_result = 1;
					  ++buff;
					  *buff = '\0';
					  --buff;
					  temp[0] = '\0';
				break;
				
				case PRINT:
					print_string( buff);
					_no_result = 1;
					*buff = '\0';
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case END:
					save_memory();
					exit(0);
				break;

				case GOTO:
					goto_line( buff);
					_no_result = 1;
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case GOSUB:
					gosub_line( buff);
					_no_result = 1;
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case RETURN:
					return_from();
					_no_result = 1;
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case WHILE:
					what_if( buff, WHILE);
					_no_result = 1;
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				
				case IF:
					what_if( buff, IF);
					_no_result = 1;
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case ENDWHILE:
					end_while();
					_no_result = 1;
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case ENDIF:
					/* do nothing - endif is more like a line label */
					/* note that you must put a colon at the end of */
					/* endif: just like at the end of a line label  */
					_no_result = 1;
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case LOAD:
					load(buff);
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case SAVE:
					save(buff);
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case SHELL:
					shell(buff);
					*buff = '\0';
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				case DUMP:
					dump();
					_no_result = 1;
					++buff;
					*buff = '\0';
					--buff;
					temp[0] = '\0';
				break;

				default:
					error( "error in function", function);
				break;

                	
			}
			strcpy(temp, buff);	/* copy the results back into the buff */
			n = 0;
			

		}
	}
	buff = buffer;
	strcpy( buff, temp);           /* return with all functions cleared */

return(0);
}

/***************************************************************************/
void dump()
{

	/********************************************************************
	 * arguments: none                                                  *
	 * returns: nothing meaningful.                                     *
	 * globals: struct _variable                                        *
	 * purpose: dumps the names and values of the variables             *
	 ********************************************************************/
	int n;

	n = 0;
	while( strlen(_variable[n].name) > 0 && n < MAX_VARS)
	{
       	if( ( _variable[n].amount < 1e10 && _variable[n].amount > 1e-9) || _variable[n].amount == 0.0)
	 		printf( "%s = %.9lf\n", _variable[n].name, _variable[n].amount);
		else
           	printf( "%s = %.9lg\n", _variable[n].name, _variable[n].amount);
    ++n;
	}


}

/***************************************************************************/
shell( buff)
char *buff;
{
	/********************************************************************
	 * arguments: string pointer buff                                   *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: execute DOS command or shell                            *
	 ********************************************************************/

	_no_result = 1;

    kill_leading_space( buff);

	if( strlen(buff) < 1)
	{
	 	printf("Type EXIT to return to BC\n\n");
		system("command.com");
		return(0);

	}
	else
		system( buff);


	return(0);

}

/***************************************************************************/

insert(buff, string, arg, start)
char *buff;
char *string;
double arg;
int start;
{
	/********************************************************************
	 * arguments: string pointers buff, string                          *
	 *            double arg                                            *
	 *            int start                                             *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: take two strings and insert a string of a double        *
	 *          in between them and return the resulting string         *
	 ********************************************************************/

	/* inserts number into string */
	char temp[MAX_BUFF + 2];

	strcpy( temp, string);

    if( ( arg < 1e10 && arg > 1e-9) || arg == 0.0)
		sprintf( &temp[ start ],"%.9lf", arg);/* format the results */
	else
		sprintf( &temp[ start ],"%.9lg", arg);/* format the results */

    remove_tail( temp);
	strcat( temp, buff);
	strcpy( buff, temp);
return(0);

}

/***************************************************************************/

what_function( function)
char *function;
{
	/********************************************************************
	 * arguments: string pointer function                               *
	 * returns: int representing what function or var held in function  *
	 * globals: none                                                    *
	 * purpose: find out what function if any has been passed to this   *
	 *          function and return its number                          *
	 ********************************************************************/

	/* table of functions - case doesn't matter */

	make_lower( function);

	if( strcmp( function, "goto") == 0)
		return( GOTO);
	if( strcmp( function, "while") == 0)
		return( WHILE);
	if( strcmp( function, "endwhile") == 0)
		return( ENDWHILE);
	if( strcmp( function, "if") == 0)
		return( IF);
	if( strcmp( function, "endif") == 0)
		return( ENDIF);
	if( strcmp( function, "gosub") == 0)
		return( GOSUB);
	if( strcmp( function, "return") == 0)
		return( RETURN);
	if( strcmp( function, "remain") == 0)
		return( REMAIN);
	if( strcmp( function, "sqr") == 0)
		return( SQR);
	if( strcmp( function, "pi") == 0)
		return( _PI);
	if( strcmp( function, "sin") == 0)
		return( SIN);
	if( strcmp( function, "cos") == 0)
		return( COS);
	if( strcmp( function, "tan") == 0)
		return( TAN);
	if( strcmp( function, "csc") == 0)
		return( CSC);
	if( strcmp( function, "sec") == 0)
		return( SEC);
	if( strcmp( function, "cot") == 0)
		return( COT);
	if( strcmp( function, "abs") == 0)
		return( ABS);
	if( strcmp( function, "acos") == 0)
		return( ACOS);
	if( strcmp( function, "asin") == 0)
		return( ASIN);
	if( strcmp( function, "atan") == 0)
		return( ATAN);
	if( strcmp( function, "ceil") == 0)
		return( CEIL);
	if( strcmp( function, "cosh") == 0)
		return( COSH);
	if( strcmp( function, "exp") == 0)
		return( EXP);
	if( strcmp( function, "ln") == 0)
		return( LN);
	if( strcmp( function, "log") == 0)
		return( LOG);
	if( strcmp( function, "rand") == 0)
		return( RAND);
	if( strcmp( function, "sinh") == 0)
		return( SINH);
	if( strcmp( function, "tanh") == 0)
		return( TANH);
	if( strcmp( function, "int") == 0)
		return( INT);
	if( strcmp( function, "cexp") == 0)
		return( CEXP);
	if( strcmp( function, "loan") == 0)
		return( LOAN);
	if( strcmp( function, "mod") == 0)
		return( MOD);
	if( strcmp( function, "hypot") == 0)
		return( HYPOT);
	if( strcmp( function, "mem") == 0)
		return( MEM);
	if( strcmp( function, "deg") == 0)
		return( DEG);
	if( strcmp( function, "rad") == 0)
		return( RAD);
	if( strcmp( function, "root") == 0)
		return( ROOT);
	if( strcmp( function, "help") == 0)
		return( HELP);
	if( strcmp( function, "fact") == 0)
		return( FACT);
	if( strcmp( function, "fract") == 0)
		return( FRACT);
	if( strcmp( function, "recip") == 0)
		return( RECIP);
	if( strcmp( function, "purm") == 0)
		return( PURM);
	if( strcmp( function, "input") == 0)
		return( INPUT);
	if( strcmp( function, "print") == 0)
		return( PRINT);
	if( strcmp( function, "end") == 0)
		return( END);
	if( strcmp( function, "save") == 0)
		return( SAVE);
	if( strcmp( function, "load") == 0)
		return( LOAD);
	if( strcmp( function, "shell") == 0)
		return( SHELL);
	if( strcmp( function, "dump") == 0)
		return( DUMP);


return( VARIABLE ); /* failed - it must be a variable */
}
/***************************************************************************/
make_lower( buff)
char *buff;
{
	/********************************************************************
	 * arguments: string pointer buff                                   *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: convert string to lower case                            *
	 ********************************************************************/

 	while( *buff)
	{
	 	*buff = tolower( *buff);
		++buff;
	}
	return(0);
}
/*************************************************************************/
gosub_line( buffer)
char *buffer;
{
	/********************************************************************
	 * arguments: string pointer buffer                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: _pc the program counter                                 *
	 * purpose: push the current line and then goto line held in buffer *
	 ********************************************************************/

	p_push(_pc); /* push the current line */
	goto_line( buffer);
	return(0);
}
/*************************************************************************/
return_from()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: _pc the program counter                                 *
	 * purpose: pops the old program posintion off the stack and resets *
	 *          the program counter to it                               *
	 ********************************************************************/

	/* pop old statement line off the stack and goto it */
	_pc = p_pop();
	return(0);
}
/*************************************************************************/
double root( index, radicand)
double index;
double radicand;
{
	/********************************************************************
	 * arguments: double index, radicand                                *
	 * returns: root index of radicand                                  *
	 * globals: none                                                    *
	 * purpose: return yth root of x                                    *
	 ********************************************************************/

	/* return x ^ 1/y extracting yth root of x */
	if( modulus( index, 2.0) ) /* if index is odd you can extract the root of a negative number */
	{
		if( radicand < 0.0)
		{
			radicand = fabs( radicand);
			return( -(pow( radicand , (1.0 / index))) );
		}
		else
			return( pow( radicand , (1.0 / index)));
	}
	return( pow( radicand , (1.0 / index)));
}

/*************************************************************************/

double integer( arg)
double( arg);
{
	/* truncate decimal fraction */
	return( (double)((long) arg) );
}
/*************************************************************************/

double loan( amount, interest, months)
double amount;
double interest;
double months;
{
	/* compute payment on a loan */

	double x, y;

	interest = interest / 1200.00;					
	x = amount * interest;							
	y = 1.0 - 1.0 / pow(( 1.0 + interest), months);	

	return( x / y);

}
/*************************************************************************/
double deg( arg)
double arg;
{
	/* convert radians to degrees */
	return( arg * 180.0 / 3.141592654 );
}
/*************************************************************************/
double rad( arg)
double arg;
{
 	/* convert degrees to radians */
	return( 3.141592654 * arg / 180 );
}
/*************************************************************************/

double cexp( arg)
double arg;
{
	/* give 10 ^ arg */
	return(pow( 10.0, arg) );
}
/*************************************************************************/

double random_number()
{
	/* I used the random number routine from the book */
	/* The Most Popular Subroutines in BASIC by Ken Tracton */
	/* returns random number range 0 - 1 */
	static double x;
	static double old_x;
	static int flag;

	if( flag != 10)
	{
		x = ( 24298.0 * (double) time(NULL) + 99991.0) / 199017.0;
		flag = 10;
	}
	else
	{
		x = (( 24298.0+(double) time(NULL) * x + 99991.0) / 1990178917.0);
	}
	if( (old_x - x) > 0.00000001)
		x = ( 24298.0 * (double) time(NULL) + 99991.0) / 199017.0;

	old_x = fract( x );

	return old_x;

}

/*************************************************************************/
double fact( n)
double n;
{
	int x;
	double f;
	f = 1.0;
	/* computes factorial */
	if( n < 0.0 )
	{
		printf("illegal factorial argument: %lf\n", n);
		_no_result = 1;
		return(0);
	}
	if( n == 0.0 )
		return( 1.0);
	for( x = 1; x <= (int) n; x++)
		f = f * (double) x;

	return f;

}
/**************************************************************************/

double fract( arg)
double arg;
{
	/* get rid of integeral part */
	return ( arg - integer( arg) );
}
/**************************************************************************/

double csc( arg)
double arg;
{
	/* cosecant function */
 	return( 1.0 / sin( arg) );
}

/**************************************************************************/

double sec( arg)
double arg;
{
 	/* secant function */
	return( 1.0 / cos( arg) );
}

/**************************************************************************/

double cot( arg)
double arg;
{
 	/* cotangent function */
	return( 1.0 / tan( arg) );

}
/**************************************************************************/

double recip( arg)
double arg;
{
	/* reciprocal */
	return (1.0 / arg);
}
/**************************************************************************/

double purm( arg1, arg2)
double arg1;
double arg2;
{
 	/* computes purmutations */
	if( arg1 < arg2)
	{
		printf( "The first argument must be equal to or greater than the second\n");
		_no_result = 1;
		return(0);
	}
	return( fact( arg1) / fact( arg1 - arg2) );

}
/***************************************************************************/

get_input( buffer)
char *buffer;
{

	/********************************************************************
	 * arguments: string pointer buffer                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: get input from keyboard or stdin after printing an      *
	 *          optional prompt string
	 ********************************************************************/


	char *temp;
	char var[ VAR_LEN + 1];
	char string[ 83];
	int n,s;
	double arg;


	n = 0;
	temp = buffer;
	temp = strchr( buffer, '\"');
	if( temp != NULL)
	{
	   ++temp;
	   while( *temp != '\"' && *temp != '\0')
	   {
			putc( *temp, stdout);      /* print out input prompt */
			++temp;
	   }
	   if( *temp == '\0')
	   		error(" error in input statement - unexpected end of string","");
	   ++temp; /* skip the quote */
	}
	else
		temp = buffer;
	while( *buffer != '\0')
	{
		n = 0;

		while ( (isspace( *temp) || *temp == ',' || *temp == ';')  && *temp != '\0')
			++temp;			/* skip spaces and commas and semicolons*/

		while ( isalpha( *temp) || *temp == '[' )
		{
			var[n] = *temp;

			if( *temp == '[')
			{
					s = 0;
					++temp;
					var[ n+1] = '\0';
					while( *temp != ']' && *temp != '\0')
					{
					 	string[s] = *temp;
						++s;
						if(  s > 78)
						{
							string[s] = '\0';
							error( " error in array subscript - subcript too long", string);
						}
						++temp;
					}
					string[s] = '\0';

					if( *temp == '\0')
						error("error in array - no closing bracket ", string);

					arg = get_result( string); /* convert the subscript to a number */
					sprintf(string,"%d", (int) arg); /* print it to the string converting it to an int */
					strcat( var, string); /* stick the subscript in there */
					strcat( var, "]");       /* stick the closing bracket on */
					n = strlen( var);
			}
			++temp;
			++n;
		}

		var[n] = '\0';

		if ( n == 0)
			error(" error in input statement - no variable found","");
		if ( strlen( var) > VAR_LEN)
			error(" error in input statement - variable too long", var);
		while( ! isalpha( *temp) &&  *temp != '\0')
			++temp;
		strcpy(buffer, temp);
		if( fgets(string, 80, stdin ) == NULL)
		{
	 		save_memory();
			exit(0);
		}

		clear_exp(string);                   /* parenthesize exponents */
		clear_paren(string);                 /* clear the parentheses */
		arg = get_result(string);  /* this allows input of expressions as well as simple digits */
		assign( var, arg);
	}
	return( 0);

}
/************************************************************************/
print_string( buffer)
char *buffer;
{
	/********************************************************************
	 * arguments: string pointer buffer                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: _no_result                                              *
	 * purpose: print strings, variables, expressions with formatting   *
	 *          punctuation. The comma tabs over to the next tab        *
	 *          position while supressing a newline, and the semicolon  *
	 *          simply supresses the newline.                           *
	 ********************************************************************/

	char *temp;
	char number[80];
	char var[ MAX_BUFF + 1];
	double result;
	int n;
	int newline = 1;

	temp = buffer;

	_no_result = 1;

	while( *temp != '\0')
	{
		n = 0;

	 	while( isspace( *temp) )
			++temp;

		if( *temp == '\"')
		{
			++temp;
		 	while( *temp != '\"' && *temp != '\0')
			{
			 	putc( *temp, stdout);
				++temp;
			}
			newline = 1;       /* yes, put a newline after the string */
			++temp;            /* skip the quote mark */
			continue;
		}
		if( *temp == ';' || *temp == ',')
		{
			while( *temp == ';' || *temp == ',')  /* is it formatting punctuation */
			{
				if( *temp == ',')
				{
					putc( '\t', stdout);
					newline = 0;
				}
				if( *temp == ';')
					newline = 0;   /* found a semicolon so don't put a newline */
				++temp;
			}
			continue;
		}

		while( *temp != ',' && *temp != ';' && *temp != '\"' && *temp != '\0')
		{
			var[n] = *temp;
			if( *temp != ' ') /* not just a space */
				newline = 1;
			if( *temp == '[')
			{
			 	get_array( temp);
			}
			++temp;
			++n;
			if( n > MAX_BUFF)
			{
				error("error in expression - too long", statement[_pc-1]);
			}
		}

		var[n] = '\0';


		kill_leading_space( var);
		if( strlen( var) < 1)
			continue;

		clear_exp( var);                   /* parenthesize exponents */
		clear_paren( var);                 /* clear the parentheses */
		result = get_result( var);

        if( ( result < 1e10 && result > 1e-9) || result == 0.0)
			sprintf( number,"%.9lf", result);/* format the results */
		else
			sprintf( number,"%.9lg", result);/* format the results */

		remove_tail( number);
		printf( " %s ", number);
	}

	if( newline)
		printf("\n");

	return( 0);

}
/************************************************************************/

kill_leading_space( buff)
char *buff;
{
	/********************************************************************
	 * arguments: string pointer buff                                   *
	 * returns: nothing meaningful.                                     *
	 * globals: none                                                    *
	 * purpose: eleminates leading blanks in string                     *
	 ********************************************************************/

 	char *p;
	p = buff;
	while( isspace( *buff))
		++buff;
	strcpy( p, buff);
return(0);
}

/************************************************************************/

goto_line( buffer)
char *buffer;
{
	/********************************************************************
	 * arguments: string pointer buffer                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: _pc                                                     *
	 * purpose: searchs for line label held in buffer and transfers     *
	 *          program execution to that line + 1
	 ********************************************************************/

	int n = 0;
	int x;
	while( isspace( *buffer))
		++buffer;
	while( isalpha( buffer[n] ) )
		++n;
	buffer[n] = '\0'; /* eleminate trailing spaces */
	if( strcmp( buffer, "endif") ==  0 || strcmp( buffer, "ENDIF") == 0 ) /* special case of going to endif */
		n = _pc;
	else 
		n = 0;
	 
	for( ; statement[n] != NULL; n++)
	{
		x = 0;
		while ( isspace( statement[n][x]) )
			++x;
		if(strncmp( &statement[n][x], buffer, strlen(buffer)) == 0)
		{
			if( strchr( statement[n], ':') )
			{
		 		_pc = n + 1; /* goto the line following the label */
				return( n);
			}
		}
	}
	error( "can't find goto label ", buffer);

return(0); /* keep the compiler warnings off my back */
}
/**************************************************************************/

what_if( buffer, operation)
char *buffer;
int operation;
{
	/********************************************************************
	 * arguments: string pointer buffer, int operation                  *
	 * returns: nothing meaningful.                                     *
	 * globals: _pc                                                     *
	 * purpose: evaluate conditional operation from an if statement or  *
	 *          a while statement and act upon it.
	 ********************************************************************/

	char oper[10];
	int n;
	char temp[80];

	n = 0;

	while ( isspace( *buffer) )
		++buffer;

	while(
			*buffer != '=' &&
			*buffer != '>' &&
			*buffer != '<' &&
			*buffer != '!' &&
			*buffer != '\0'
		 )
	{
		temp[n] = *buffer;
		++n;
		++buffer;
		if( *buffer == '\0')
			error( "error in if statement ", statement[_pc-1]);
	}
	temp[n] = '\0';
	get_number( temp);

	while( isspace( *buffer))
		++buffer;

	n = 0;

	while( *buffer == '=' || *buffer == '>' || *buffer == '<' || *buffer == '!')
	{
		oper[n] = *buffer;
		++n;
		if( *buffer == '\0')
			error( "error in statement ", statement[_pc -1]);
		++buffer;	
		if( n > 9)
			error(" error in relational operator","");
	}
	oper[n] = '\0';

	get_number( buffer);

	if( strcmp( oper, "!=") == 0)
	{
		if( pop() != pop())
		{
			if( operation == WHILE)
			{
				p_push( _pc - 1);
			}
			return(0); 
		}
		else
		{
            	end_if( operation);
			return(1);
		}
	}
	if( strcmp( oper, "=") == 0)
	{
		if( pop() == pop() )
		{
			if( operation == WHILE)
			{
				p_push( _pc - 1);
			}
			return(0);
		}
		else
		{
            end_if( operation);
			return(1);
		}

	}
	if( strcmp( oper, "<=") == 0)
	{
		swap();
		if( pop() <= pop() )
		{
			if( operation == WHILE)
			{
				p_push( _pc - 1);
			}
			return(0);
		}
		else
		{
            end_if( operation);
			return(1);
		}

	}

	if( strcmp( oper, ">=") == 0)
	{
		swap();
		if( pop() >= pop() )
		{
			if( operation == WHILE)
			{
				p_push( _pc - 1);
			}
			return(0);
		}
		else
		{
            end_if( operation);
			return(1);
		}

	}

	if( strcmp( oper, ">") == 0)
	{
		swap();
		if( pop() > pop() )
		{
			if( operation == WHILE)
			{
				p_push( _pc - 1);
			}
			return(0);
		}
		else
		{
            end_if( operation);
			return(1);
		}

	}

	if( strcmp( oper, "<") == 0)
	{
		swap();
		if( pop() < pop() )
		{
			if( operation == WHILE)
			{
				p_push( _pc - 1);
			}
			return(0);
		}
		else
		{
            end_if( operation);
			return(1);
		}

	}

	error( "error in if statement operator ", statement[_pc -1]);

return(0); /* TurboC keeps carping at me so here is a worthless return */

}
/***************************************************************************/
end_while()
{
	/********************************************************************
	 * arguments: none.                                                 *
	 * returns: nothing meaningful.                                     *
	 * globals: _pc                                                     *
	 * purpose: go back to the while statement indicated by the program *
	 *          stack                                                   *
	 ********************************************************************/

	_pc = p_pop();
	return(0);
}
/***************************************************************************/
end_if( operation)
int operation;
{
	/********************************************************************
	 * arguments: int operation                                         *
	 * returns: nothing meaningful.                                     *
	 * globals: _pc                                                     *
	 * purpose: finds the matching endif or endwhile when the condition *
	 *          in the statement evaluates to false.                    *
	 ********************************************************************/

 	int if_level = 1; /* to match if to matching endif */

	while( statement[_pc] != NULL)
	{
	 	if( isaloop( statement[_pc]) || isaif( statement[_pc]) )
		{
		 	/* if we find a if and and we don't find a quote mark increment the if levels */
			if ( operation == IF)
			{
				if( ! strstr( statement[_pc], "endif") && ! strstr( statement[_pc], "ENDIF") )
				{
					if( strstr( statement[_pc], "if") || strstr( statement[_pc], "IF") )
					{
						++_pc;
						++if_level;
						continue;
					}
				}
			}

		}
		if( operation == IF)
		{
			if( (strstr( statement[_pc], "ENDIF") ||  strstr( statement[_pc], "endif")) && ! strchr( statement[_pc], '\"') )
			{
		 		/* if we find a endif not in quotes check out the if_level */
				if( if_level == 1 )
				{
					++_pc;
			 		return(0); /* found the matching endif */
				}
				--if_level;

			}
		}
		if( operation == WHILE)
		{
			if( (strstr( statement[_pc], "ENDWHILE") ||  strstr( statement[_pc], "endwhile")) && ! strchr( statement[_pc], '\"') )
			{
		 		/* if we find a endwhile not in quotes check out the if_level */
				if( if_level == 1 )
				{
					++_pc;
			 		return(0); /* found the matching endwhile */
				}
				--if_level;

			}
			if( ! strstr( statement[_pc], "endwhile") && ! strstr( statement[_pc], "ENDWHILE") )
			{
				if( strstr( statement[_pc], "while") || strstr( statement[_pc], "WHILE"))
				{
					++_pc;
					++if_level;
					continue;
				}
			}

		}
		++_pc; /* increment the program counter */
	}
	error( "can't find the matching endif","");
return(1);

}
/***************************************************************************/
isaloop( buff)
char *buff;
{
	/********************************************************************
	 * arguments: string pointer buff                                   *
	 * returns: true or false                                           *
	 * globals: none                                                    *
	 * purpose: compares strings to "while" and returns comparison      *
	 ********************************************************************/

	if( (strstr( buff, "WHILE") || strstr( buff, "while")) && ! strchr( buff, '\"') )
 		return(1);
	else
 		return(0);

}
/***************************************************************************/
isaif( buff)
char *buff;
{
	/********************************************************************
	 * arguments: string pointer buff                                   *
	 * returns: true or false                                           *
	 * globals: none                                                    *
	 * purpose: see if string contains the word "if"                    *
	 ********************************************************************/

	if( (strstr( buff, "IF") || strstr( buff, "if")) && ! strchr( buff, '\"') )
		return(1);
	else
		return(0);
}
/***************************************************************************/

save(fname)
char *fname;
{
	/********************************************************************
	 * arguments: string pointer name                                   *
	 * returns: nothing meaningful.                                     *
	 * globals: memory                                                  *
	 * purpose: saves all variables to disk. If it can't it sets memory *
	 *          to false and prints an error message
	 ********************************************************************/

	/* saves all the variables in memory in file <name>
	   if there is an error the routine aborts and places a
	   0 in the memory varible. else it places a 1 there. */
	FILE *fp;
	int n;

	_no_result = 1;

	while( isspace( *fname) )
		++fname;

	if( (fp = fopen( fname, "w") ) == NULL)
	{
		printf("couldn't open %s\n", fname);
		memory = 0;
		return(0);
	}
	else
	{
		n = 0;
		while( strlen(_variable[n].name) > 0 && n < MAX_VARS)
		{
        	if( ( _variable[n].amount < 1e10 && _variable[n].amount > 1e-9) || _variable[n].amount == 0.0)
		 		fprintf( fp, "%s=%.9lf\n", _variable[n].name, _variable[n].amount);
			else
            	fprintf( fp, "%s=%.9lg\n", _variable[n].name, _variable[n].amount);

			++n;
		}
	}

 	fclose(fp);
	memory = 1; /* success */

	return(0);

}
/***************************************************************************/

load(name)
char *name;
{
	/********************************************************************
	 * arguments: string pointer name                                   *
	 * returns: nothing meaningful.                                     *
	 * globals: memory                                                  *
	 * purpose: loads a file holding the variable table. If succesful,  *
	 *          places a 1 in the global double memory, else a -1       *
	 ********************************************************************/

	/* loads all the variables to memory from file <name>
	   if there is an error the routine aborts and places a
	   0 in the memory varible. else it places a 1 there. */

	FILE *fp;
	char buffer[ MAX_BUFF + 1];
	double amount;

	_no_result = 1;

	while( isspace( *name) )
		++name;

	if( (fp = fopen( name, "r") ) == NULL)
	{
		printf("couldn't open %s\n", name);
		memory = 0;
		return(0);
	}
	else
	{
		while( fgets( buffer, MAX_BUFF, fp) != NULL)
		{
			clear_exp( buffer);
			clear_paren( buffer);
			_no_result = 1;
			amount = get_result( buffer);

		}
	}
 	fclose(fp);
	memory = 1; /* success */

	return(0);

}
/***************************************************************************/
double modulus( arg1, arg2)
double arg1;
double arg2;
{
 	double x;
	if( arg1 == 0) return(0);

	x = arg1 / arg2;
	if( x > 0.0)
		x = ( fract( x) * arg2);
	else
	{
		x =  fract( x ) * (arg2 - 1);
		x = arg2 - fabs(x);
		x = integer(x);
	}

	return(fabs(x));
}

/***************************************************************************/

double remain( arg1, arg2)
double arg1;
double arg2;
{
 	double x;
	if( arg1 == 0) return(0);

	x = arg1 / arg2;
		x = ( fract( x) * arg2);

	return(x);
}
